require 'date'
require 'json'
require 'forwardable'

#===============================================================================
# Xeme
#

# Objects of this class represent a set of results. See the README file for
# more details.

class Xeme
	# version 0.3.1
	VERSION = '0.3.1'
	
	# A Xeme::Messages object, which is basically just a hash containing arrays
	# for errors, warnings, and notes.
	attr_reader :messages
	
	# A hash of any miscellaneous details you want to include.
	attr_reader :misc
	
	#---------------------------------------------------------------------------
	# initialize
	#
	
	# new() does not take any parameters.
	
	def initialize
		@messages = Xeme::Messages.new
		@misc = {}
		@prefixes = []
		@auto_details_val = nil
		@transaction = nil
	end
	#
	# initialize
	#---------------------------------------------------------------------------
	
	
	#---------------------------------------------------------------------------
	# errors, warnings, notes
	#
	
	##
	# Returns the array of errors.
	def errors
		return @messages['errors']
	end
	
	##
	# Returns the array of warnings.
	def warnings
		return @messages['warnings']
	end
	
	##
	# Returns the array of notes.
	def notes
		return @messages['notes']
	end
	#
	# errors, warnings, notes
	#---------------------------------------------------------------------------
	
	
	#---------------------------------------------------------------------------
	# transaction
	#
	
	##
	# Returns the transaction object, creating it if necessary. See
	# Xeme::Transaction. If the transaction object is never created then it is
	# not output with #to_json. The transaction object can be used to provide
	# meta information about the request and response that produced the
	# results. See Xeme::Transaction for more information.
	#
	# You can create the transaction object by simply calling this method:
	#
	#	xeme.transaction
	#
	# If you want to assign a request ID, call this method and assign to its
	# request property. For example, if there's a request object that provides
	# an id, you could assign the transaction.request ID like this:
	#
	#	xeme.transaction.request = request.id
	
	def transaction
		@transaction ||= Xeme::Transaction.new()
		return @transaction
	end
	#
	# transaction
	#---------------------------------------------------------------------------
	
	
	#---------------------------------------------------------------------------
	# to_h
	#
	
	##
	# Returns a hash with all the results. This hash can be used to create a JSON
	# object. Empty arrays, such as if there are no warnings, are not included.
	
	def to_h
		# $tm.hrm
		
		# initialize return value
		rv = {}
		
		# success
		rv['success'] = success()
		
		# transaction
		if @transaction
			rv['transaction'] = @transaction.to_h
		end
		
		# if any messages
		if @messages.any?
			msgs = rv['messages'] = {}
			
			@messages.each do |key, arr|
				if arr.any?
					msgs[key] = []
					
					arr.each do |msg|
						msgs[key].push msg.to_h
					end
				end
			end
		end
		
		# misc
		if @misc.any?
			rv['misc'] = @misc
		end
		
		# return
		return rv
	end
	#
	# to_h
	#---------------------------------------------------------------------------
	
	
	#---------------------------------------------------------------------------
	# to_json
	#
	
	##
	# Returns a JSON string containing all the result information.
	def to_json(opts={})
		if opts['pretty']
			return JSON.pretty_generate(self.to_h)
		else
			return JSON.generate(self.to_h)
		end
	end
	#
	# to_json
	#---------------------------------------------------------------------------
	
	
	#---------------------------------------------------------------------------
	# success
	#
	
	##
	# Returns false if there are any errors, true otherwise.
	
	def success		
		return @messages['errors'].empty?
	end
	
	##
	# Returns the opposite of <tt>success()</tt>.
	def failure
		return success ? false : true
	end
	
	#
	# success
	#---------------------------------------------------------------------------
	
	
	#---------------------------------------------------------------------------
	# save_exception
	#
	
	# Use this method to hold on to details about an exception. An error message
	# object will be created, along with the exception's to_s and backtrace.
	
	def save_exception(id, e, details={})
		message = self.error(id, details)
		message['error'] = e.to_s
		message['backtrace'] = e.backtrace
	end
	#
	# save_exception
	#---------------------------------------------------------------------------
	
	
	#---------------------------------------------------------------------------
	# prefix
	#
	
	##
	# <tt>prefix</tt> allows you to automatically add prefixes to your message
	# codes in situations where many messages will start with the same string.
	# Consider the following situation.
	#
	# 	xeme.error '/virginia/slope'
	# 	xeme.error '/virginia/angle'
	# 	xeme.error '/virginia/departure'
	#
	# That works, but the redundancy can get confusing (and annoying). With
	# <tt>prefix</tt> you can avoid the redundancy like this:
	#
	#	xeme.prefix('virginia') do
	#	    xeme.error 'slope'
	#	    xeme.error 'angle'
	#	    xeme.error 'departure'
	#	end
	#
	# 	puts xeme.errors[0].id # /virginia/slope
	# 	puts xeme.errors[1].id # /virginia/angle
	# 	puts xeme.errors[2].id # /virginia/departure
	#
	# Prefixes can also be nested, like this:
	#
	# 	xeme.prefix('virginia') do
	#	    xeme.prefix('fairfax') do
	#	        xeme.error 'slope'
	#	    end
	# 	end
	#
	# 	puts xeme.errors[0].id # /virginia/fairfax/slope
	
	def prefix(p_prefix)
		# $tm.hrm
		hold_prefixes = @prefixes.clone
		@prefixes = @prefixes.clone
		@prefixes.push p_prefix
			
		begin
			yield
		ensure
			@prefixes = hold_prefixes
		end
	end
	#
	# prefix
	#---------------------------------------------------------------------------
	
	
	#---------------------------------------------------------------------------
	# error, warning, note
	#
	
	##
	# Creates a message object and stores it in the errors array.
	#
	#	xeme.error('error-1').
	#
	# A hash of arbitrary details can be given. Those details will be stored
	# with the message object in its Xeme::Message#details property.
	#
	#	xeme.error('error-0', 'choice'=>'a')
	#
	# This method returns the message object, which can be used as another
	# way to store details.
	#
	#	msg = xeme.error('error-1').
	#	msg['location'] = 1
	#
	# If a block is given, the block is called with the error object as the
	# single param.
	#
	#	xeme.error('error-2') do |error|
	#	    error['location'] = 2
	#	end
	def error(id, details={}, &block)
		return add_comment(Xeme::Message, 'errors', id, details={}, &block)
	end
	
	##
	# Works just #error() except it stores the message in the warnings array.
	def warning(id, details={}, &block)
		return add_comment(Xeme::Message, 'warnings', id, details={}, &block)
	end
	
	##
	# Works just #error() except it stores the message in the notes array.
	def note(id, details={}, &block)
		return add_comment(Xeme::Message, 'notes', id, details={}, &block)
	end
	#
	# error, warning, note
	#---------------------------------------------------------------------------
	
	
	#---------------------------------------------------------------------------
	# from_json
	#
	
	# Creates a new Xeme object from a JSON structure.
	
	def self.from_json(raw_json)
		# $tm.hrm
		hsh = JSON.parse(raw_json)
		rv = self.new
		
		# messages
		if messages = hsh['messages']
			messages.each do |msg_type, arr|
				rv.messages[msg_type] ||= []
				rv.messages[msg_type] += arr
			end
		end
		
		# misc
		if misc = hsh['misc']
			misc.each do |key, val|
				rv.misc[key] = val
			end
		end
		
		# transaction
		if transaction = hsh['transaction']
			if transaction['timestamp']
				rv.transaction.timestamp = DateTime.parse(transaction['timestamp'])
			end
			
			if transaction['request']
				rv.transaction.request = transaction['request']
			end
			
			if transaction['response']
				rv.transaction.response = transaction['response']
			end
		end
		
		# return
		return rv
	end
	#
	# from_json
	#---------------------------------------------------------------------------
	
	
	#---------------------------------------------------------------------------
	# new_exception
	#
	
	# Creates and returns a Xeme::Exception object. The Xeme object is attached
	# to the Xeme::Exception object.
	
	def new_exception(error_id)
		return Xeme::Exception.new(error_id, self)
	end
	#
	# new_exception
	#---------------------------------------------------------------------------
	
	
	# private methods
	private

	
	#---------------------------------------------------------------------------
	# messages_to_h
	#
	def messages_to_h(arr)
		rv = []
		
		# messages
		arr.each do |msg|
			rv.push msg.to_h
		end
		
		# return
		return rv
	end
	#
	# messages_to_h
	#---------------------------------------------------------------------------
	
	
	#---------------------------------------------------------------------------
	# add_comment
	#
	def add_comment(cls, type, id, details={}, &block)
		# $tm.hrm
		
		# build message
		msg = cls.new(prefix_id(id), details)
		
		# ensure array
		@messages[type] ||= []
		
		# add to errors array
		@messages[type].push(msg)
		
		# yield
		if block_given?
			yield msg
		end
		
		# return
		return msg
	end
	#
	# add_comment
	#---------------------------------------------------------------------------
	
	
	#---------------------------------------------------------------------------
	# prefix_id
	#
	def prefix_id(p_id)
		if @prefixes.any?
			return '/' + @prefixes.join('/') + '/' + p_id
		else
			return p_id
		end
	end
	#
	# prefix_id
	#---------------------------------------------------------------------------
end
#
# Xeme
#===============================================================================


#===============================================================================
# Xeme::Message
#

# A Xeme::Message object represents a single error, warning, or note.
# It is the base class for Xeme::Message::Error,
# Xeme::Message::Warning, and 
# Xeme::Message::Note. Don't instantiate Xeme::Message
# directly, use Xeme#error, Xeme#warning, or Xeme#note.
#
# Each message contains at least an id. It can also have an options details
# hash which can be used to store misc information about the message.
#
# The message itself can be treated like a hash to give details:
#
#	msg = xeme.error('my-error')
#	msg['location'] = 1

class Xeme::Message
	# The id of the message.
	attr_reader :id
	
	# Details about the message.
	attr_reader :details
	
	# delegate to hsh
	extend Forwardable
	delegate %w([] []= each length clear delete) => :@details
	
	
	#---------------------------------------------------------------------------
	# initialize
	#
	
	# Initialize a Xeme::Message object with the id of the message. The id can
	# be any string. The optional <tt>details</tt> hash will be stored with the
	# message.
	
	def initialize(id, details={})
		# $tm.hr __method__
		
		# set id
		@id = id
		
		# initialize details
		@details = details.clone
	end
	#
	# initialize
	#---------------------------------------------------------------------------
	
	
	#---------------------------------------------------------------------------
	# to_h
	#
	
	##
	# Returns a hash structure of the message. This structure is used by
	# Xeme#to_h.
	
	def to_h
		rv = {}
		rv['id'] = @id
		
		if @details.any?
			rv['details'] = @details
		end
		
		return rv
	end
	#
	# to_h
	#---------------------------------------------------------------------------
end
#
# Xeme::Message
#===============================================================================


#===============================================================================
# Xeme::Messages
#

##
# This object holds the arrays of errors, warnings, and notes. It can mostly be
# treated like a hash. The main difference between a Xeme::Messages object and
# a hash is that any? returns true or false based on if there are any messages,
# not if there are any elements in the hash.

class Xeme::Messages
	# delegate to hsh
	extend Forwardable
	delegate %w([] []= each length keys has_key?) => :@hsh
	
	#---------------------------------------------------------------------------
	# initialize
	#
	
	##
	# new() takes no parameters.
	
	def initialize
		@hsh = {}
		@hsh['errors'] = []
		@hsh['warnings'] = []
		@hsh['notes'] = []
	end
	#
	# initialize
	#---------------------------------------------------------------------------
	
	
	#---------------------------------------------------------------------------
	# any?
	#
	
	##
	# Returns if there are any messages in any of the errors, warnings, or
	# notes arrays.
	
	def any?
		@hsh.values.each do |arr|
			arr.any? and return true
		end
		
		return false
	end
	#
	# any?
	#---------------------------------------------------------------------------
end
#
# Xeme::Messages
#===============================================================================


#===============================================================================
# Xeme::Transaction
#

# An object of this class provides meta information about the request and
# results. It always provides a timestamp and a unique ID for the results. It
# may also optionally include a request ID that was provided by the process
# that made the request, such as a call to a REST application. Do not directly
# instantiate this class; use Xeme#transaction.

class Xeme::Transaction
	
	# Optional. An ID for the request that was sent by the calling process. Do
	# not generate this value yourself; use the ID that was sent with the
	# request (if one was sent).
	attr_accessor :request
	
	# Gives a unique ID for these results.
	attr_accessor :response
	
	# Gives a timestamp for when these results were generated.
	attr_accessor :timestamp
	
	#---------------------------------------------------------------------------
	# initialize
	#
	
	# Initialize does not take any parameters.
	
	def initialize
		@timestamp = DateTime.now()
		@response = rand().to_s.sub(/\A0\./mu, '')
		@request = nil
	end
	#
	# initialize
	#---------------------------------------------------------------------------
	
	
	#---------------------------------------------------------------------------
	# to_h
	#
	def to_h
		rv = {}
		
		# request
		if @request
			rv['request'] = @request
		end
		
		# response and timestamp
		rv['response'] = @response
		rv['timestamp'] = @timestamp.to_s
		
		# return
		return rv
	end
	#
	# to_h
	#---------------------------------------------------------------------------
end
#
# Xeme::Transaction
#===============================================================================


#===============================================================================
# Xeme::Exception
#

# Xeme::Exception is a subclass of StandardError and can be used exactly like a
# StandardError object. The one addition to StandardError is the `xeme` property
# which holds a Xeme object. Don't instantiate this class directly; use
# Xeme#new_exception.

class Xeme::Exception < StandardError
	attr_reader :xeme # a xeme object
	
	#---------------------------------------------------------------------------
	# initialize
	#
	def initialize(error_code, xeme)
		@error_code = error_code
		@xeme = xeme
		super @error_code
	end
	#
	# initialize
	#---------------------------------------------------------------------------
end
#
# Xeme::Exception
#===============================================================================
